import base64
import json
from calendar import timegm
from datetime import datetime, UTC, timedelta
from textwrap import wrap

from cryptography import x509
from cryptography.hazmat._oid import NameOID
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric.rsa import generate_private_key
from cryptography.hazmat.primitives.serialization import load_pem_private_key, load_pem_public_key, Encoding
from dateutil.relativedelta import relativedelta
from jose import jwk, jws
from jose.constants import ALGORITHMS

"""
Any variables prefixed with `NV_` or `nv_` are original values, dumped from an NLS registration/activation process.
Any variables prefixed with `MY_` or `my_` are variables which are set by fastapi-dls and are "faked".
"""

### FILES

FILE_REQUEST_ROUTING_SI = '../../doc/database/3-after-upload-license/request_routing.service_instance.json'
FILE_CONFIG_TOKEN = '../../doc/files/config-token.json'
FILE_SI_ARTIFACT = '../../doc/database/3-after-upload-license/si_d8c07e4af6a449d0b2dc3faf0e1bf2bd.service_instance_artifact.json'
FILE_SI_CONFIG = '../../doc/database/3-after-upload-license/si_d8c07e4af6a449d0b2dc3faf0e1bf2bd.configuration.json'

### DEFAULTS

with open(FILE_REQUEST_ROUTING_SI, 'r') as f:
    NV_SI_SITE_ID = json.loads(f.read())[0].get('xid')

with open(FILE_CONFIG_TOKEN, 'r') as f:
    NV_CONFIG_TOKEN_RESPONSE = json.loads(f.read())

with open(FILE_SI_ARTIFACT, 'r') as f:
    si_artifact_rows = json.loads(f.read())
    si_identity_rows = list(filter(lambda _: _.get('namespace') == 'service_instance.client.all', si_artifact_rows))
    si_identity_private_key = next(filter(lambda _: _.get('name') == 'private_key', si_identity_rows))
    si_identity_public_key = next(filter(lambda _: _.get('name') == 'public_key', si_identity_rows))
    NV_SI_KEY_RSA = si_identity_private_key.get('value')
    NV_SI_KEY_PUB = si_identity_public_key.get('value')

MY_CLIENT_TOKEN_EXPIRE_DELTA = relativedelta(years=12)

### VARIABLES

nv_si_private_key = load_pem_private_key(NV_SI_KEY_RSA.encode('utf-8'), password=None)
nv_si_private_key_pem = nv_si_private_key.private_bytes(
    encoding=serialization.Encoding.PEM,
    format=serialization.PrivateFormat.TraditionalOpenSSL,
    encryption_algorithm=serialization.NoEncryption()
)
nv_si_public_key = load_pem_public_key(NV_SI_KEY_PUB.encode('utf-8'))
nv_si_public_key_pem = nv_si_public_key.public_bytes(
    encoding=serialization.Encoding.PEM,
    format=serialization.PublicFormat.SubjectPublicKeyInfo
)
nv_si_public_key_exp = nv_si_public_key.public_numbers().e
nv_si_public_key_mod = f'{nv_si_public_key.public_numbers().n:x}'  # hex value without "0x" prefix


### TESTS

def test_private_public_key_equals_public_key():
    """
    test that the public-key exported from private-key is the same as the defined public-key

    :return:
    """

    nv_public_key_from_private_key_as_pem = nv_si_private_key.public_key().public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo
    )
    nv_public_key_as_pem = nv_si_public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo
    )
    assert nv_public_key_as_pem == nv_public_key_from_private_key_as_pem


def test_official_cert():
    """
    test the right certificate

    :return:
    """

    nv_response_certificate_configuration = NV_CONFIG_TOKEN_RESPONSE.get('certificateConfiguration')
    nv_response_ca_cert = nv_response_certificate_configuration.get('caChain')[0].encode('utf-8')
    nv_response_public_cert = nv_response_certificate_configuration.get('publicCert').encode('utf-8')
    nv_response_public_key = nv_response_certificate_configuration.get('publicKey')

    nv_response_parsed_ca_cert = x509.load_pem_x509_certificate(nv_response_ca_cert)
    nv_response_parsed_cert = x509.load_pem_x509_certificate(nv_response_public_cert)

    assert 4096 == nv_response_parsed_ca_cert.public_key().key_size
    assert 2048 == nv_response_parsed_cert.public_key().key_size
    #assert nv_response_parsed_cert.public_key().public_bytes(encoding=serialization.Encoding.PEM, format=serialization.PublicFormat.SubjectPublicKeyInfo) == nv_si_public_key.public_bytes(encoding=serialization.Encoding.PEM, format=serialization.PublicFormat.SubjectPublicKeyInfo)

    nv_response_parsed_cert_exp = nv_response_parsed_cert.public_key().public_numbers().e
    nv_response_parsed_cert_mod = f'{nv_response_parsed_cert.public_key().public_numbers().n:x}'  # hex value without "0x" prefix
    assert nv_response_parsed_cert_exp == nv_response_public_key.get('exp')
    assert nv_response_parsed_cert_mod == nv_response_public_key.get('mod')[0]

    nv_issuer = f'CN=NLS Intermediate CA,OU=Nvidia Licensing Service (NLS),O=Nvidia,ST=California,C=US'
    nv_subject = f'CN={NV_SI_SITE_ID}'
    assert nv_response_parsed_cert.issuer.rfc4514_string() == nv_issuer
    assert nv_response_parsed_cert.subject.rfc4514_string() == nv_subject

    nv_ca_ski = nv_response_parsed_ca_cert.extensions.get_extension_for_class(x509.SubjectKeyIdentifier).value
    nv_cert_aki = nv_response_parsed_cert.extensions.get_extension_for_class(x509.AuthorityKeyIdentifier).value
    assert nv_ca_ski.key_identifier == nv_cert_aki.key_identifier


def test_official_config_token_response():
    """
    test config-token response and test jwt-signing

    :return:
    """

    nv_response_certificate_configuration = NV_CONFIG_TOKEN_RESPONSE.get('certificateConfiguration')
    nv_response_public_cert = nv_response_certificate_configuration.get('publicCert').encode('utf-8')
    nv_jwt_decode_key = jwk.construct(nv_response_public_cert, algorithm=ALGORITHMS.RS256)

    nv_response_config_token = NV_CONFIG_TOKEN_RESPONSE.get('configToken')

    payload = jws.verify(nv_response_config_token, key=nv_jwt_decode_key, algorithms=ALGORITHMS.RS256)
    payload = json.loads(payload)
    assert payload.get('iss') == 'NLS Service Instance'
    assert payload.get('aud') == 'NLS Licensed Client'
    assert payload.get('service_instance_ref') == NV_SI_SITE_ID

    nv_si_public_key_configuration = payload.get('service_instance_public_key_configuration')
    nv_si_public_key_me = nv_si_public_key_configuration.get('service_instance_public_key_me')
    assert nv_si_public_key_me.get('mod') == nv_si_public_key_mod
    assert nv_si_public_key_me.get('exp') == nv_si_public_key_exp


def test_our_config_token():
    """
    test our config-token with nvidia data and test jwt-signing.

    THIS TEST WILL FAIL UNTIL WE HAVE FOUND THE CORRECT PRIVATE KEY

    :return:
    """

    nv_response_certificate_configuration = NV_CONFIG_TOKEN_RESPONSE.get('certificateConfiguration')
    nv_response_public_cert = nv_response_certificate_configuration.get('publicCert').encode('utf-8')
    nv_response_parsed_cert = x509.load_pem_x509_certificate(nv_response_public_cert)

    nv_response_parsed_cert_as_pem = nv_response_parsed_cert.public_bytes(encoding=Encoding.PEM)
    jwt_decode_key = jwk.construct(nv_response_parsed_cert_as_pem, algorithm=ALGORITHMS.RS256)

    ### build payload

    cur_time = datetime.now(UTC)
    exp_time = cur_time + MY_CLIENT_TOKEN_EXPIRE_DELTA

    my_payload = {
        "iss": "NLS Service Instance",
        "aud": "NLS Licensed Client",
        "iat": timegm(cur_time.timetuple()),
        "nbf": timegm(cur_time.timetuple()),
        "exp": timegm(exp_time.timetuple()),
        "protocol_version": "2.0",
        "d_name": "DLS",
        "service_instance_ref": NV_SI_SITE_ID,
        "service_instance_public_key_configuration": {
            "service_instance_public_key_me": {
                "mod": hex(nv_si_public_key.public_numbers().n)[2:],
                "exp": int(nv_si_public_key.public_numbers().e),
            },
            "service_instance_public_key_pem": nv_si_public_key_pem.decode('utf-8'),
            "key_retention_mode": "LATEST_ONLY"
        },
    }

    # todo: maybe DLS_SI_CERTIFICATE['private_key'], but how to decrypt?!
    # our_correct_sign_key = load_key('where_is_our_correct_private_key.pem').export_key().decode('utf-8')
    # our_correct_sign_key = jwk.construct(our_correct_sign_key, algorithm=ALGORITHMS.RS256)
    # our_correct_config_token = jws.sign(payload, key=our_correct_sign_key, headers=None, algorithm=ALGORITHMS.RS256)
    # fails:
    # - Table: "service_instance_artifact" => "service_instance.client.all" => "private_key"
    # - Table: "service_instance_artifact" => "service_instance.identity" => "private_key"
    # - Table: "public_private_key_pair" => "private_key"
    # this will fail, until we have not found the correct private key
    # "jwt_encode_key" has invalid signature (can't be verified with DLS_SI_CERTIFICATE['certificate'])
    with open(FILE_SI_CONFIG, 'r') as f:
        rows = json.loads(f.read())
        dls_si_certificate = next(filter(lambda _: _.get('property_name') == 'DLS_SI_CERTIFICATE', rows))
        dls_si_certificate = dls_si_certificate.get('property_value')
        dls_si_certificate_private_key = dls_si_certificate.get('private_key')
        dls_si_certificate_private_key = base64.b64decode(dls_si_certificate_private_key)
    # Mengsk @ discord:
    # I think it's AES-GCM encrypted, from the cert I saw key length is 2048b,
    # which should be 1732 bytes in pem format.
    # private_key is 1744 bytes looks like pem + 12 bytes gcm
    assert 1744 == len(dls_si_certificate_private_key)
    # So that this does not work currently, we'll use code below to have a "working" test example.
    # In the future, this lines will replace the "placeholder" below
    # my_sign_key = jwk.construct(dls_si_certificate_private_key.decode('utf-8'), algorithm=ALGORITHMS.RS256)
    # my_config_token = jws.sign(my_payload, key=my_sign_key, headers=None, algorithm=ALGORITHMS.RS256)
    placeholder_sign_key = jwk.construct(nv_si_private_key_pem, algorithm=ALGORITHMS.RS256)

    my_config_token = jws.sign(my_payload, key=placeholder_sign_key, headers=None, algorithm=ALGORITHMS.RS256)

    my_response = {
        "certificateConfiguration": {
            "caChain": NV_CONFIG_TOKEN_RESPONSE['certificateConfiguration']['caChain'],
            "publicCert": NV_CONFIG_TOKEN_RESPONSE['certificateConfiguration']['publicCert'],
            "publicKey": {
                "exp": int(nv_response_parsed_cert.public_key().public_numbers().e),
                "mod": [hex(nv_response_parsed_cert.public_key().public_numbers().n)[2:]],
            },
        },
        "configToken": my_config_token,
    }

    ###

    my_config_token = my_response.get('configToken')

    # todo: this "jws.verify" will fail, because we *sign* with "nv_si_private_key_pem" but *verify* with
    # the certificate (which dont belong together)
    payload = jws.verify(my_config_token, key=jwt_decode_key, algorithms=ALGORITHMS.RS256)
    payload = json.loads(payload)
    assert payload.get('iss') == 'NLS Service Instance'
    assert payload.get('aud') == 'NLS Licensed Client'
    assert payload.get('service_instance_ref') == NV_SI_SITE_ID

    service_si_public_key_configuration = payload.get('service_instance_public_key_configuration')
    service_si_public_key_me = service_si_public_key_configuration.get('service_instance_public_key_me')
    assert service_si_public_key_me.get('mod') == nv_si_public_key_mod
    assert service_si_public_key_me.get('exp') == nv_si_public_key_exp


def test_our_config_token_with_our_key():
    """
    test our config-token and test jwt-signing

    :return:
    """

    """ Create Root Key and Certificate """

    # create root keypair
    my_root_private_key = generate_private_key(public_exponent=65537, key_size=4096)
    my_root_public_key = my_root_private_key.public_key()

    # create root-certificate subject
    my_root_subject = x509.Name([
        x509.NameAttribute(NameOID.COUNTRY_NAME, u'US'),
        x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, u'California'),
        x509.NameAttribute(NameOID.ORGANIZATION_NAME, u'Nvidia'),
        x509.NameAttribute(NameOID.ORGANIZATIONAL_UNIT_NAME, u'Nvidia Licensing Service (NLS)'),
        x509.NameAttribute(NameOID.COMMON_NAME, u'NLS Root CA'),
    ])

    # create self-signed root-certificate
    my_root_certificate = (
        x509.CertificateBuilder()
        .subject_name(my_root_subject)
        .issuer_name(my_root_subject)
        .public_key(my_root_public_key)
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.now(tz=UTC) - timedelta(days=1))
        .not_valid_after(datetime.now(tz=UTC) + timedelta(days=365 * 10))
        .add_extension(x509.BasicConstraints(ca=True, path_length=None), critical=True)
        .add_extension(x509.SubjectKeyIdentifier.from_public_key(my_root_public_key), critical=False)
        .sign(my_root_private_key, hashes.SHA256()))

    """ Create CA (Intermediate) Key and Certificate """

    # create ca keypair
    my_ca_private_key = generate_private_key(public_exponent=65537, key_size=4096)
    my_ca_public_key = my_ca_private_key.public_key()

    # create ca-certificate subject
    my_ca_subject = x509.Name([
        x509.NameAttribute(NameOID.COUNTRY_NAME, u'US'),
        x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, u'California'),
        x509.NameAttribute(NameOID.ORGANIZATION_NAME, u'Nvidia'),
        x509.NameAttribute(NameOID.ORGANIZATIONAL_UNIT_NAME, u'Nvidia Licensing Service (NLS)'),
        x509.NameAttribute(NameOID.COMMON_NAME, u'NLS Intermediate CA'),
    ])

    # create self-signed ca-certificate
    my_ca_certificate = (
        x509.CertificateBuilder()
        .subject_name(my_ca_subject)
        .issuer_name(my_root_subject)
        .public_key(my_ca_public_key)
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.now(tz=UTC) - timedelta(days=1))
        .not_valid_after(datetime.now(tz=UTC) + timedelta(days=365 * 10))
        .add_extension(x509.BasicConstraints(ca=True, path_length=None), critical=True)
        .add_extension(x509.KeyUsage(digital_signature=False, key_encipherment=False, key_cert_sign=True,
                                     key_agreement=False, content_commitment=False, data_encipherment=False,
                                     crl_sign=True, encipher_only=False, decipher_only=False), critical=True)
        .add_extension(x509.SubjectKeyIdentifier.from_public_key(my_ca_public_key), critical=False)
        # .add_extension(x509.AuthorityKeyIdentifier.from_issuer_public_key(my_root_public_key), critical=False)
        .add_extension(x509.AuthorityKeyIdentifier.from_issuer_subject_key_identifier(
            my_root_certificate.extensions.get_extension_for_class(x509.SubjectKeyIdentifier).value
        ), critical=False)
        .sign(my_root_private_key, hashes.SHA256()))

    with open('caChain_my.pem', 'wb') as f:
        f.write(my_ca_certificate.public_bytes(encoding=Encoding.PEM))

    """ Create Service-Instance Key and Certificate """

    # create si keypair
    my_si_private_key = generate_private_key(public_exponent=65537, key_size=2048)
    my_si_public_key = my_si_private_key.public_key()

    my_si_private_key_as_pem = my_si_private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption(),
    )
    my_si_public_key_as_pem = my_si_public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    )

    with open('instance.private.pem', 'wb') as f:
        f.write(my_si_private_key_as_pem)

    with open('instance.public.pem', 'wb') as f:
        f.write(my_si_public_key_as_pem)

    # create si-certificate subject
    my_si_subject = x509.Name([
        x509.NameAttribute(NameOID.COMMON_NAME, u'4e53a171-103b-4946-9ed8-5f4c0ee750d9'),
    ])

    # create self-signed si-certificate
    my_si_certificate = (
        x509.CertificateBuilder()
        .subject_name(my_si_subject)
        .issuer_name(my_ca_subject)
        .public_key(my_si_public_key)
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.now(tz=UTC) - timedelta(days=1))
        .not_valid_after(datetime.now(tz=UTC) + timedelta(days=365 * 10))
        .add_extension(x509.KeyUsage(digital_signature=True, key_encipherment=True, key_cert_sign=False,
                                     key_agreement=True, content_commitment=False, data_encipherment=False,
                                     crl_sign=False, encipher_only=False, decipher_only=False), critical=True)
        .add_extension(x509.ExtendedKeyUsage([
            x509.oid.ExtendedKeyUsageOID.SERVER_AUTH,
            x509.oid.ExtendedKeyUsageOID.CLIENT_AUTH]
        ), critical=False)
        .add_extension(x509.SubjectKeyIdentifier.from_public_key(my_si_public_key), critical=False)
        # .add_extension(x509.AuthorityKeyIdentifier.from_issuer_public_key(my_ca_public_key), critical=False)
        .add_extension(x509.AuthorityKeyIdentifier.from_issuer_subject_key_identifier(
            my_ca_certificate.extensions.get_extension_for_class(x509.SubjectKeyIdentifier).value
        ), critical=False)
        .add_extension(x509.SubjectAlternativeName([
            x509.DNSName('4e53a171-103b-4946-9ed8-5f4c0ee750d9')
        ]), critical=False)
        .sign(my_ca_private_key, hashes.SHA256()))

    my_si_public_key_exp = my_si_certificate.public_key().public_numbers().e
    my_si_public_key_mod = f'{my_si_certificate.public_key().public_numbers().n:x}'  # hex value without "0x" prefix

    with open('cert_my.pem', 'wb') as f:
        f.write(my_si_certificate.public_bytes(encoding=Encoding.PEM))

    """ build out payload """

    cur_time = datetime.now(UTC)
    exp_time = cur_time + MY_CLIENT_TOKEN_EXPIRE_DELTA

    payload = {
        "iss": "NLS Service Instance",
        "aud": "NLS Licensed Client",
        "iat": timegm(cur_time.timetuple()),
        "nbf": timegm(cur_time.timetuple()),
        "exp": timegm(exp_time.timetuple()),
        "protocol_version": "2.0",
        "d_name": "DLS",
        "service_instance_ref": NV_SI_SITE_ID,
        "service_instance_public_key_configuration": {
            "service_instance_public_key_me": {
                "mod": hex(my_si_public_key.public_numbers().n)[2:],
                "exp": int(my_si_public_key.public_numbers().e),
            },
            # 64 chars per line (pem default)
            "service_instance_public_key_pem": my_si_public_key_as_pem.decode('utf-8').strip(),
            "key_retention_mode": "LATEST_ONLY"
        },
    }

    my_sign_key = jwk.construct(my_si_private_key_as_pem.decode('utf-8'), algorithm=ALGORITHMS.RS256)
    my_config_token = jws.sign(payload, key=my_sign_key, headers=None, algorithm=ALGORITHMS.RS256)

    # generate a 76 char wide pem (does not work either, so this code can be removed)
    response_ca_chain = my_ca_certificate.public_bytes(encoding=Encoding.PEM).decode('utf-8')
    response_ca_chain = response_ca_chain.replace('-----BEGIN CERTIFICATE-----', '')
    response_ca_chain = response_ca_chain.replace('-----END CERTIFICATE-----', '')
    response_ca_chain = response_ca_chain.replace('\n', '')
    response_ca_chain = wrap(response_ca_chain, 76)
    response_ca_chain = '\r\n'.join(response_ca_chain)
    response_ca_chain = f'-----BEGIN CERTIFICATE-----\r\n{response_ca_chain}\r\n-----END CERTIFICATE-----'

    response_si_certificate = my_si_certificate.public_bytes(encoding=Encoding.PEM).decode('utf-8')
    response_si_certificate = response_si_certificate.replace('-----BEGIN CERTIFICATE-----', '')
    response_si_certificate = response_si_certificate.replace('-----END CERTIFICATE-----', '')
    response_si_certificate = response_si_certificate.replace('\n', '')
    response_si_certificate = wrap(response_si_certificate, 76)
    response_si_certificate = '\r\n'.join(response_si_certificate)
    response_si_certificate = f'-----BEGIN CERTIFICATE-----\r\n{response_si_certificate}\r\n-----END CERTIFICATE-----'

    response = {
        "certificateConfiguration": {
            # 76 chars per line
            "caChain": [response_ca_chain],
            # 76 chars per line
            "publicCert": response_si_certificate,
            "publicKey": {
                "exp": int(my_si_certificate.public_key().public_numbers().e),
                "mod": [hex(my_si_certificate.public_key().public_numbers().n)[2:]],
            },
        },
        "configToken": my_config_token,
    }

    # this would be our response on fastapi-dls
    print(json.dumps(response))

    #my_si_certificate_as_pem = my_si_certificate.public_bytes(encoding=Encoding.PEM)
    #my_jws_verify_key = jwk.construct(my_si_certificate_as_pem, algorithm=ALGORITHMS.RS256)
    my_response_certificate = x509.load_pem_x509_certificate(response['certificateConfiguration']['publicCert'].encode('utf-8'))
    my_response_certificate_as_pem = my_response_certificate.public_bytes(encoding=Encoding.PEM)
    my_jws_verify_key = jwk.construct(my_response_certificate_as_pem, algorithm=ALGORITHMS.RS256)

    my_response_config_token = response.get('configToken')

    payload = jws.verify(my_response_config_token, key=my_jws_verify_key, algorithms=ALGORITHMS.RS256)
    payload = json.loads(payload)
    assert payload.get('iss') == 'NLS Service Instance'
    assert payload.get('aud') == 'NLS Licensed Client'
    assert payload.get('service_instance_ref') == NV_SI_SITE_ID

    my_si_public_key_configuration = payload.get('service_instance_public_key_configuration')
    my_si_public_key_me = my_si_public_key_configuration.get('service_instance_public_key_me')
    assert my_si_public_key_me.get('mod') == my_si_public_key_mod
    assert my_si_public_key_me.get('exp') == my_si_public_key_exp
